#!/usr/bin/perl -w 

use strict;
use CGI;
use File::Find;
use Fcntl ':flock'; # import LOCK_* constants

print "Content-type: text/html\n\n";

my $query =  new CGI;

my $password_file = "/home/virtuals/tad/etc/htpasswd";
my $group_file = "/home/virtuals/tad/etc/htgroup";
my $htdocs_dir = "/home/virtuals/tad/public_html";
my $cgi_dir = "/home/virtuals/tad/public_html/cgi-bin";

# I could check for all groups by searching for .htaccess files.
# I won't bother for the moment. (Update: this script does now)

sub groups_of_user {
 my $user = shift;
 open(GROUP_FILE,"$group_file") || return ();
 flock( GROUP_FILE, LOCK_EX );
 my @groups = ();
 while (<GROUP_FILE>) {
    next unless /^(\w+):.*\b$user\b/;
    @groups = (@groups,$1);
 }
 close(GROUP_FILE);
 flock( GROUP_FILE, LOCK_UN );
 return @groups;
}

sub users_of_group {
 my $group = shift;
 open(GROUP_FILE,"$group_file") || return ();
 flock( GROUP_FILE, LOCK_EX );
 my @users = ();
 while (<GROUP_FILE>) {
    next unless /^$group:\s*(.*)\s*$/;
    @users = (@users,split(/\s+/,$1));
 }
 close(GROUP_FILE);
 flock( GROUP_FILE, LOCK_UN );
 return @users;

}

my @htfiles = ();
sub htaccess_files { 
  if ($_ eq ".htaccess") { @htfiles = (@htfiles,$File::Find::dir."/".$_); }
}

my @known_groups = ();
sub find_all_groups {
  if ($#known_groups != -1) { return; }
  @htfiles = ();
  &File::Find::find( \&htaccess_files, $htdocs_dir, $cgi_dir ); # set up @htfiles
  my $htfile;
  foreach $htfile (@htfiles) { 
    open(HTFILE,"$htfile") || next; 
    while (<HTFILE>) {
      next unless /^\s*require group (\w+)/;
      next if grep ($_ eq $1,@known_groups);
      @known_groups = (@known_groups,$1);
    }
    close(HTFILE);
  }
  open(GROUP_FILE,"$group_file") || die "Can't read $group_file";
  flock( GROUP_FILE, LOCK_EX );
  my $group;
  while (<GROUP_FILE>) {
    next unless /^(\w+):/;
#/;  emacs food
    $group = $1;
    next if (grep($group eq $_,@known_groups)); 
    @known_groups = (@known_groups,$group);
  }
  close(GROUP_FILE);
  flock( GROUP_FILE, LOCK_UN );
}


sub set_user_to_have_these_groups {
 my ($user) = shift;
 my (@groups) = @_; 
 my ($group); 
 my ($in_old_list);
 my ($in_new_list);
 my (%seen) = ();
 my ($previous_line_leftovers) = "";
 my ($previous_group) = "";
    my (%groups_already_written) = ();
 open(GROUP_FILE,"$group_file") || die "Can't read group file";
 flock( GROUP_FILE, LOCK_EX );
 open(NEW_GROUP_FILE,">/tmp/grp.$$") || die "Can't write to group temp file";
 # The only complication here is that the group might be broken out into
 # several lines.
 while (<GROUP_FILE>) {
   if ($_ !~ /^(\w+):/) {
     print NEW_GROUP_FILE $_;  next;
   }
   $group = $1;
   if (($previous_group ne $group) and ($previous_line_leftovers ne "")) {
       # if the previous line was so big that even the leftovers were more
       # than 8192 characters, then we have a problem. However, it will be
       # fixed the next time anyone in this line gets referenced again.
       print NEW_GROUP_FILE "$previous_group: $previous_line_leftovers\n";
       $previous_line_leftovers = "";
   }
   $in_new_list = grep($_ eq $group,@groups) ? 1 : 0;
   $in_old_list = /\b$user\b/;
   if ($in_new_list and $in_old_list) {  }
   if ($in_new_list and $in_old_list and exists($groups_already_written{$group})) {
       # then we would have already written it
       s/\b$user\b//g;
   }
   if ($in_new_list and !$in_old_list and !exists($seen{$group})) {
       # we only want it written there once -- the first line, hopefully.
       $groups_already_written{$group} = 1;
       chomp; $_ .= " $user\n"; 
   }
   if (!$in_new_list and $in_old_list) { s/\b$user\b//g; }
   if (!$in_new_list and !$in_old_list) { } 
   if ($previous_group eq $group) { 
       chomp; $_ .= " $previous_line_leftovers\n";
       $previous_line_leftovers = "";
   }
   s/ {2,}/ /g; 
   if ($_ =~ /(.{1000,}?)\b +\b(.*)/) {
       # line too long!
       $_ = "$1\n";
       $previous_line_leftovers = $2;
       chomp $previous_line_leftovers;
   }
   print NEW_GROUP_FILE "$_";
   if ($in_new_list) { $seen{$group} = 1; }
   $previous_group = $group;
 }
 if ($previous_line_leftovers ne "") {
     print NEW_GROUP_FILE "$previous_group: $previous_line_leftovers\n";
     $previous_line_leftovers = "";
 }
 foreach $group (@groups) { 
   next if (defined $seen{$group}) ;
   print NEW_GROUP_FILE "$group: $user\n";  # put users into never-seen-before groups
 }
 close (NEW_GROUP_FILE);
 close (GROUP_FILE);
 system ("mv /tmp/grp.$$ $group_file");
 flock( GROUP_FILE, LOCK_UN );
}

sub create_new_group {
 &find_all_groups();
 my ($group) = shift;
 if (grep($_ eq $group,@known_groups)) { return; }
 open(GROUP_FILE,">>$group_file") || die "Can't read group file";
 flock( GROUP_FILE, LOCK_EX );
 print GROUP_FILE "$group:\n";
 close(GROUP_FILE);
 @known_groups = (@known_groups,$group);
}

sub delete_group {
 my ($group) = shift;
 open(GROUP_FILE,"$group_file") || die "Can't read group file";
 flock( GROUP_FILE, LOCK_EX );
 open(NEW_GROUP_FILE,">/tmp/grp.$$") || die "Can't write to group temp file";
 while (<GROUP_FILE>) {
   next if /^$group:/;
   print NEW_GROUP_FILE "$_";
 }
 close (NEW_GROUP_FILE);
 close (GROUP_FILE);
 system ("mv /tmp/grp.$$ $group_file");
 flock( GROUP_FILE, LOCK_UN );
 @known_groups = ();
}


my @all_users = ();
sub find_all_users {
 if ($#all_users != -1) { return; }
 open(PWFILE,$password_file) || die ("Can't open $password_file on ".`hostname`." from ".`pwd`." as uid = $< , $> and gid = $(, $) because $!");
 flock( PWFILE, LOCK_EX );
 while (<PWFILE>) {
  next unless /^([^:]*):([^:]*)/;
  @all_users = (@all_users,$1);
 }
 close(PWFILE);
 flock(PWFILE, LOCK_UN );
}

#sub encrypt_password {
# my $new_password = shift;
# my $salt = join '', ('.', '/', 0..9, 'A'..'Z', 'a'..'z')[rand 64, rand 64];
# my $encrypted = crypt($new_password,$salt);
# return $encrypted;
#}

sub add_new_user {
 &find_all_users();
 my ($username, $password) = @_;
 if (grep($_ eq $username,@all_users)) { &set_password($username,$password); return; }
 #my $encrypted = &encrypt_password($password);
 #open(PWFILE,">>$password_file") || die "Can't open $password_file";
 #flock( PWFILE, LOCK_EX );
 #print PWFILE "$username:$encrypted\n";
 #close (PWFILE);
 $password =~ s/\\/\\\\/;
 $password =~ s/'/\'/;
 system("/usr/local/bin/htpasswd -b $password_file $username '$password' 2> /dev/null");
 @all_users = (@all_users,$username);
}

sub delete_user {
 my ($username) = @_;
 open(PWFILE,"$password_file") || die "Can't open $password_file";
 flock( PWFILE, LOCK_EX );
 open(PWFILE2,">/tmp/pwfile.$$") || die "Can't create tmp-reopening nasty temp file";
 while (<PWFILE>) {
  next if (/^$username:/);
  print PWFILE2 $_; 
 }
 close(PWFILE);
 close(PWFILE2);
 system("mv /tmp/pwfile.$$ $password_file"); 
 flock(PWFILE, LOCK_UN );
 &set_user_to_have_these_groups($username);
 @all_users = ();
}
 

sub set_password {
 my ($username, $password) = @_;
 #my $encrypted = &encrypt_password($password);
 $password =~ s/\\/\\\\/;
 $password =~ s/'/\'/;
 system("/usr/local/bin/htpasswd -b $password_file $username '$password' 2> /dev/null");
 #open(PWFILE,"$password_file") || die "Can't open $password_file";
 #flock( PWFILE, LOCK_EX );
 #open(PWFILE2,">/tmp/pwfile.$$") || die "Can't create tmp-reopening nasty temp file";
 #while (<PWFILE>) {
 # if (/^$username:/) { print PWFILE2 "$username:$encrypted\n"; } 
 # else { print PWFILE2 $_; }
 #}
 #close(PWFILE);
 #close(PWFILE2);
 #system("mv /tmp/pwfile.$$ $password_file"); 
 #flock(PWFILE, LOCK_UN );
}


sub error_out {
 my $problem = $_[0];
 print "<HTML><HEAD><TITLE>Error</TITLE></HEAD><BODY BGCOLOR=WHITE>\n";
 print "I was unable to do that because: <BR>\n";
 print "<FONT COLOR=RED>$problem</RED></FONT>\n";
 print "</HTML>\n";
 exit(0);
}

if (defined ($query->param('username'))) {
  my $username = $query->param('username');
  &error_out("Bad characters in $username") unless $username =~ /^[a-zA-Z0-9_-]+$/;

  if (defined ($query->param('groups')) or defined ($query->param('purpose') and $query->param('purpose') eq 'modify-groups')) {
   # Then we are updating the groups for the user.
   my @groups = $query->param('groups');
   my @bad_groups = grep ($_ !~ /^[a-zA-Z0-9_-]+$/,@groups);
   if ($#bad_groups != -1) { &error_out("Bad characters in the following groups: ". join(" ",@bad_groups)); }
   &set_user_to_have_these_groups($username,@groups);
   &display_user($username);
 }
  
 if (defined ($query->param('confirm-delete'))) {
   &delete_user($username);
   &display_all_users();
 }

 if (defined ($query->param('password'))) {
   &add_new_user($username,$query->param('password'));
   &display_user($username);
 }

 # otherwise,  no task given...
 &display_user($username);
}
 

if (defined ($query->param('groupname'))) {
 my $groupname = $query->param('groupname');
 &error_out("$groupname has bad characters in it") unless $groupname =~ /^[a-zA-Z0-9_-]+$/;
 if (defined ($query->param('confirm-delete'))) {
   &delete_group($groupname);
   &display_all_groups();
 } 
 if (defined ($query->param('confirm-create'))) {
  &create_new_group($groupname);
  &display_group($groupname);
 }
 &display_group($groupname);
}

if (defined ($query->param('show'))) {
 my $what = $query->param('show');
 if ($what eq "groups") { &display_all_groups(); }
 if ($what eq "users") { &display_all_users(); }
}

&display_all_users();

sub header {
 print "<HTML><HEAD><TITLE>User management</TITLE></HEAD><BODY BGCOLOR=WHITE>";
 print "\n";
}

sub footer {
 print "<HR><h2>Other functions...</h2>\n";
 print "<FORM METHOD=POST>";
 print "Create a new user: <br><br>\n";
 print "<b>Username</b>: <INPUT TYPE=TEXT NAME=username><br>\n";
 print "<b>Password</b>: <INPUT TYPE=PASSWORD NAME=password>\n";
 print "<INPUT TYPE=SUBMIT VALUE=\"Create user\">\n";
 print "</FORM>\n";
 print "<HR><SMALL><UL>\n";
 print '<li><a href="manage-users.pl?show=users"> Complete list of users </a></li>'."\n";
 print '<li><a href="manage-users.pl?show=groups"> Complete list of groups </a></li>'."\n";
 print "</ul></small></body></html>\n";
}

sub display_all_users {
 &header();
 print "<H1>User list</H1>\n";
 &find_all_users();
 my $user;
 my $i=0;
 my @breakpoints = (0,int (($#all_users+1)/3),
                      int (2*($#all_users+1)/3)
                      );
  print "<!-- ".join(", ",@breakpoints). "-->\n";
 print "<table><tr>\n";
 foreach  $user (sort @all_users) { 
   print "<td><ul>" if (grep($_ == $i,@breakpoints));   $i++;
   print "<li><a href=\"manage-users.pl?username=$user\"> $user </a></li>\n";
   print "</ul></td>" if (grep($_ == $i,@breakpoints)); 
 }
 print "</tr></table>\n";
 &footer();
 exit(0);
}

sub display_user {
 my $username = shift;
 &header();
 print "<h1>$username</h1>\n";
 my @groups = &groups_of_user($username);
 my $group;
 &find_all_groups();
 print "<h2>Group Membership</h2>\n";
 print "<form method=post>\n";
 my $check;
 foreach $group (@known_groups) {
   $check = grep ($group eq $_, @groups) ? "CHECKED=CHECKED" : "";
   print "<INPUT TYPE=CHECKBOX $check NAME=groups VALUE=\"$group\">\n";
   print "$group<Br>\n";
 }
 print "<P>\n";
 print "<INPUT TYPE=RESET>\n";
 print "<INPUT TYPE=SUBMIT VALUE=\"Update\"><bR>\n";
 print "<INPUT TYPE=HIDDEN NAME=purpose VALUE=\"modify-groups\">\n";
 print "<INPUT TYPE=HIDDEN NAME=username VALUE=\"$username\">\n";
 print "</FORM>\n";
 print "<hr>\n";
 print "<H2>Fix forgotten password</h2>\n";
 print "<FORM METHOD=POST>\n";
 print "<INPUT TYPE=HIDDEN NAME=username VALUE=\"$username\">\n";
 print "<b>Password:</b> <INPUT TYPE=PASSWORD NAME=password><br>\n";
 print "<INPUT TYPE=SUBMIT VALUE=\"Set password\"><bR>\n";
 print "</FORM>\n";
 print "<hr>\n";
 print "<h2>Delete user</h2>\n";
 print "<FORM METHOD=POST>\n";
 print "<INPUT TYPE=HIDDEN NAME=\"confirm-delete\"  VALUE=1>\n";
 print "<INPUT TYPE=HIDDEN NAME=username VALUE=\"$username\">\n";
 print "<INPUT TYPE=SUBMIT VALUE=\"Delete user\"><bR>\n";
 print "</FORM>\n";
 &footer();
 exit(0);
}



sub display_all_groups {
 &find_all_groups ();
 my $group;
 &header();
 print "<h1>Groups list</h1>\n";
 print "<ul>\n";
 foreach $group (@known_groups) {
  print "<li><a href=\"manage-users.pl?groupname=$group\"> $group </a></li>\n";
 }
 print "</ul>\n";
 &footer();
 exit(0);
}

sub display_group {
 my $group = shift;
 &header();
 print "<H1>$group</h1>\n";
 print "<h2>Members of the group</h2>\n";
 print "<ul>\n";
 my @users = &users_of_group ($group);
 my $user;
 foreach $user (@users) {  
  print " <li> <a href=\"manage-users.pl?username=$user\"> $user </a> </li> \n";
 }
 print "</ul>\n";
 if ($group ne "admin") { 
  print "<hr>\n";
  print "<h2>Clear this group</h2>\n";
  print "Clear this group of all members,  and if possible, delete it.\n";
  print "<FORM METHOD=POST>\n";
  print "<INPUT TYPE=HIDDEN NAME=groupname VALUE=\"$group\"> \n";
  print "<INPUT TYPE=HIDDEN NAME=confirm-delete VALUE=1>\n";
  print "<INPUT TYPE=SUBMIT VALUE=\"Delete $group\">\n";
  print "</FORM>\n";
 }
 print "<hr><h2>Create a new group</h2>\n";
 print "Note that there is no need to ever create a group manually.\n";
 print "The only groups needed here are found by looking through the \n";
 print "<tt>.htaccess</tt> files.\n";
 
 &footer();
 exit(0);
}



#######################################################################
# Regression tests...
#my $username = $query->param('username');
#if (!defined ($username)) {
#my @x = &groups_of_user("bakerg");
#print "@x\n";

#&find_all_groups();
#print "@known_groups\n";

#&find_all_users();
#print "@all_users\n";

#&set_password("bakerg","foobar");
#&set_user_

#
#print "Creating user smithj\n";
#&add_new_user("smithj","foobar");
#system("more $password_file");
#
#print "Creating groups users1 and users2\n";
#&create_new_group("users1");
#&create_new_group("users2");
#system("more $group_file");
#
#print "Putting smithj in users1 and users2\n";
#&set_user_to_have_these_groups("smithj","users1","users2");
#system("more $group_file");
#
#print "Taking smithj out of users2\n";
#&set_user_to_have_these_groups("smithj","users1");
#system("more $group_file");
#
#print "Putting smithj back into users2\n";
#&set_user_to_have_these_groups("smithj","users1","users2");
#system("more $group_file");
#
#print "Taking smithj out of users1\n";
#&set_user_to_have_these_groups("smithj","users2");
#system("more $group_file");
#
#print "Deleting smithj\n";
#&delete_user("smithj");
#system("more $password_file");
#system("more $group_file");
#
#print "Deleting users1\n";
#&delete_group("users1");
#system("more $group_file");
#
#print "Adding users2 a second time\n";
#&create_new_group("users2");
#system("more $group_file");
#
#print "Deleting users2\n";
#&delete_group("users2");
#system("more $group_file");
##
