#!/usr/bin/perl -w 

my $password_file = "/home/virtuals/tad/etc/htpasswd";
my $group_file = "/home/virtuals/tad/etc/htgroup";
my $htdocs_dir = "/home/virtuals/tad/public_html";
my $cgi_dir = "/home/virtuals/tad/public_html/cgi-bin";
my $htdocs_url = "http://www.technicalaidnsw.org.au";
my $cgibin_url = "http://www.technicalaidnsw.org.au/cgi-bin";

use strict;
use CGI;
use File::Find;
use File::Basename;
use Fcntl ':flock'; # import LOCK_* constants

print "Content-type: text/html\n\n";

my $query =  new CGI;
my $appname = $ENV{'SCRIPT_FILENAME'};

# I could check for all groups by searching for .htaccess files.
# I won't bother for the moment. (Update: this script does now)

sub groups_of_user {
 my $user = shift;
 open(GROUP_FILE,"$group_file") || return ();
 flock( GROUP_FILE, LOCK_EX );
 my @groups = ();
 while (<GROUP_FILE>) {
    next unless /^(\w+):.*\b$user\b/;
    @groups = (@groups,$1);
 }
 flock( GROUP_FILE, LOCK_UN );
 close(GROUP_FILE);
 return @groups;
}


my @htfiles = ();
sub htaccess_files { 
  if ($_ eq ".htaccess") { @htfiles = (@htfiles,$File::Find::dir."/".$_); }
}

sub read_htfile {
  my $htfile = shift;
  my @interesting_groups = @_;
  open(HTFILE,"$htfile") || next; 
  my $realm = undef;
  my $directory = File::Basename::dirname($htfile);
  my $url = $directory; 
       $url =~ s:$htdocs_dir:$htdocs_url:g;
       $url =~ s:$cgi_dir:$cgibin_url:g;
       $url =~ s:cgi-bin/admin:admin/manage-users.pl:g;
  my $use_this = 0; 
  my $allowed_groups;
  my @allowed_groups;
  my $allowed_group;
  while (<HTFILE>) {
    if (/^\s*require valid-user/) { $use_this = 1; next; }
    if (/^\s*AuthName\s*"(.*)"/) { $realm = $1; next; }
    next unless /^\s*require group (.+)/;
    $allowed_groups = $1;  
    $allowed_groups =~ s/^\s*//; $allowed_groups =~ s/\s*$//;
    @allowed_groups = split(/\s+/,$allowed_groups);
    foreach $allowed_group (@allowed_groups) {
       if (grep ($_ eq $allowed_group,@interesting_groups)) {
	  $use_this = 1;
          next;
       }
    }
  }
  close(HTFILE);
  return undef unless $use_this;
  return undef unless defined($realm);
  return undef if File::Basename::dirname($appname) eq $directory;
  return { 'realm' => $realm, 'url' => $url, 'dir' => $directory };
}

sub find_urls_for_groups {
  my @groups = @_;
  @htfiles = ();
  &File::Find::find( \&htaccess_files, $htdocs_dir, $cgi_dir ); # set up @htfiles
  my $htfile;
  my %urls = ();
  my $x;
  my $dir;
  foreach $htfile (@htfiles) { 
    $x = read_htfile($htfile,@groups);
    if (defined($x)) { 
       $dir = $x->{'dir'};
       $urls{$dir} = $x;
    }
  }
  return \%urls;
}



sub error_out {
 my $problem = $_[0];
 print "<HTML><HEAD><TITLE>Error</TITLE></HEAD><BODY BGCOLOR=WHITE>\n";
 print "I was unable to do that because: <BR>\n";
 print "<FONT COLOR=RED>$problem</RED></FONT>\n";
 print "</HTML>\n";
 exit(0);
}

my $username = $ENV{'REMOTE_USER'};
my $nice_username = $username; $nice_username =~ s/_/ /g;
my @groups = groups_of_user($username);
my $urls = find_urls_for_groups(@groups);
my $url;
my $substitution = "";
foreach $url (values %$urls) {
  $substitution .= "<li> <a href=\"". $url->{"url"} . "\">" . $url->{"realm"} . 
      "</a></li>\n"; 
}

if (!  open(APPEARANCE,"login-appearance.html") ) {
 &header();
 print $substitution;
 &footer();
} else {
 while(<APPEARANCE>) {
      s/LIST GOES HERE/$substitution/g;
      print;
 }
 close (APPEARANCE);
}

sub header {
 print "<HTML><HEAD><TITLE>Folders available for $nice_username</TITLE></HEAD><BODY BGCOLOR=WHITE>";
 print "<H1>Folders available for $nice_username</H1>\n";
 print "\n<ul>";
}

sub footer {
 print "</UL><HR></BODY></HTML>\n";
}


